/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2006/06/02   Fujitsu)Yositake      First Version
 */

package org.naregi.pse.acs.client;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;
import org.apache.log4j.Logger;

import org.naregi.pse.acs.client.databean.*;
import org.naregi.pse.acs.client.jaxb.AAD;
import org.naregi.pse.acs.client.jaxb.AAIDType;
import org.naregi.pse.acs.client.jaxb.AuthorType;
import org.naregi.pse.acs.client.jaxb.ContentType;
import org.naregi.pse.acs.client.jaxb.ContentsType;
import org.naregi.pse.acs.client.jaxb.DescriptionType;
import org.naregi.pse.acs.client.jaxb.DescriptionsType;
import org.naregi.pse.acs.client.jaxb.DiffAAIDType;
import org.naregi.pse.acs.client.jaxb.DiffContentType;
import org.naregi.pse.acs.client.jaxb.DiffContentsType;
import org.naregi.pse.acs.client.jaxb.DifferentialAAD;
import org.naregi.pse.acs.client.jaxb.ObjectFactory;
import org.naregi.pse.acs.client.jaxb.AuthorType.LocationType;
import org.naregi.pse.acs.client.namespace.AADNamespacePrefixMapperImpl;
import org.naregi.pse.ap.common.PSESystemInfo;

/**
 * This class provides functionalities to create AAFile.zip.
 */
public class AADFileFactory {

	private static Logger log = Logger.getLogger(AADFileFactory.class.getName());
	
	public static final String AAD_FILENAME = "aad.xml";
	
	private static final String AAID_NAME_URI = "http:/naregi/pse/acs/";
	private static final String TYPE_NAMESPACE = "http://www.naregi.org/acs/2006/04/type";
	private static final String TYPE_NAMESPACE_PREFIX = "nrg";
	
	public static final String ACS_CLIENT_TEMP_BASEDIR_KEY = "ACS_CLIENT_TEMP_BASEDIR";
	public static final int CREATE = 0;
	public static final int UPDATE = 1;

	private String aadTempDir;
	private String aadFile;
	private AcsAppInfo appInfo;
	private AcsFileInfo[] fileInfo;
	private ObjectFactory objFactory;

	/**
	 * Default Constructor
	 */
	public AADFileFactory() {

		// read from property file
		this.aadTempDir =
			PSESystemInfo.getInstance().getProperty(ACS_CLIENT_TEMP_BASEDIR_KEY);

		if (this.aadTempDir.endsWith("/") == false) {
			this.aadTempDir = this.aadTempDir + "/";
		}
		this.aadTempDir = this.aadTempDir + "AAFileFactory";
	}

	/**
	 * Constructor
	 */
	public AADFileFactory(String _aadTempDir) {

		this.aadTempDir = _aadTempDir;
	}

	public void createAAD(
		AcsAppInfo _appInfo,
		AcsFileInfo[] _fileInfo,
		int operationType)
		throws Throwable {

		// set to object properties
		this.appInfo = _appInfo;
		this.fileInfo = _fileInfo;
		
		// make sure dir exists
		makeSureDirExists(this.aadTempDir);
		
		// make sure file exists
		makeSureFileExists(this.aadTempDir, AAD_FILENAME);
		
		this.aadFile = new File(this.aadTempDir).getPath() + "/" + AAD_FILENAME;

		// create aad.xml in AADTemp location
		if(operationType == CREATE){
			createAADFile();
		}else if(operationType == UPDATE){
			createDiffAADFile();
		}
	}

	private void makeSureDirExists(String _dir) throws Exception{

		File dir = new File(_dir);
		if (dir.exists() == false) {
			if (dir.mkdirs() == false) {
				try {
					throw new Exception(dir.getPath() + ": mkdir failed");
				} catch (Exception e) {
					log.fatal(dir.getPath() + ": mkdir failed", e);
					throw e;
				}
			}
		}		
	}

	private void makeSureFileExists(String _dir, String _file) throws Exception{
		
		File dir = new File(_dir);
		File file = new File(dir.getPath() + "/" + _file);
		if(file.exists() == false){
			if(file.createNewFile() == false){
				try {
					throw new Exception(file.getPath() + ": create file failed");
				} catch (Exception e) {
					log.fatal(file.getPath() + ": create file failed", e);
					throw e;
				}				
			}
		}
	}

	public void createAADFile()
		throws Throwable {

		JAXBContext jaxbContext = null;
		ArrayList fileList = new ArrayList();

		// create JAXBContext
		jaxbContext = JAXBContext.newInstance("org.naregi.pse.acs.client.jaxb");

		// create Marshaller
		Marshaller marshaller = jaxbContext.createMarshaller();

		// create an ObjectFactory instance.
		this.objFactory = new ObjectFactory();

		// create AAD
		AAD aad = objFactory.createAAD();

		// create AAID
		aad.setAAID(
			createAAID(
				appInfo.getAppName(),
				appInfo.getAppVersion()));

		// create author
		aad.setAuthor(
			createAuthor(
				appInfo.getAuthorName(),
				appInfo.getAuthorDesc(),
				appInfo.getAuthorCountry(),
				appInfo.getAuthorAddress()));

		// create description
		if(appInfo.getAppDesc() != null){
			aad.setDescriptions(
				createDescriptions(appInfo.getAppDesc()));
		}

		// create contents
		aad.setContents((ContentsType)createContents(fileInfo));

		// set property for namespace prefix
		marshaller.setProperty("com.sun.xml.bind.namespacePrefixMapper", new AADNamespacePrefixMapperImpl());		
		
		// set property for formatted output
		marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);

		// Marshal to aadFile
		try {
			marshaller.marshal(aad, new FileOutputStream(new File(aadFile)));
		} catch (Throwable t) {
			log.fatal("Failed to marshal.", t);
			throw t;
		}
	}

	public void createDiffAADFile() throws Throwable{
			
		JAXBContext jaxbContext = null;
		ArrayList fileList = new ArrayList();

		// create JAXBContext
		jaxbContext = JAXBContext.newInstance("org.naregi.pse.acs.client.jaxb");

		// create Marshaller
		Marshaller marshaller = jaxbContext.createMarshaller();

		// create an ObjectFactory instance.
		this.objFactory = new ObjectFactory();

		// create DiffAAD
		DifferentialAAD diffAAD = objFactory.createDifferentialAAD();

		// create DiffAAID
		diffAAD.setAAID(
			createDiffAAID(
				appInfo.getAppName(),
				appInfo.getAppVersion(),
				appInfo.getAppBaseVersion()));

		// create author
		if(appInfo.getAuthorName() != null){
	
			diffAAD.setAuthor(
				createAuthor(
					appInfo.getAuthorName(),
					appInfo.getAuthorDesc(),
					appInfo.getAuthorCountry(),
					appInfo.getAuthorAddress()));			
		}

		// create description
		if(appInfo.getAppDesc() != null){
			diffAAD.setDescriptions(
				createDescriptions(appInfo.getAppDesc()));
		}

		// create contents
		diffAAD.setContents(createDiffContents(fileInfo));

		// set property for namespace prefix
		marshaller.setProperty("com.sun.xml.bind.namespacePrefixMapper", new AADNamespacePrefixMapperImpl());

		// set property for formatted output
		marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);

		// Marshal to aadFile
		try {
			marshaller.marshal(diffAAD, new FileOutputStream(new File(aadFile)));
		} catch (Throwable t) {
			log.fatal("Failed to marshal.", t);
			throw t;
		}			
	}

	private AAIDType createAAID(
		String _name,
		String _version)
		throws Exception {

		if(_name == null){
			throw new Exception("Application name required.");
		}
		
		if(_version == null){
			throw new Exception("Application version required.");
		}
		
		AAIDType aaID = objFactory.createAAIDType();

		String name = AAID_NAME_URI + _name;
		aaID.setName(name);
		aaID.setVersion(_version);
		return aaID;
	}

	private DiffAAIDType createDiffAAID(
		String _name,
		String _version,
		String _baseVersion)
		throws Exception {

		if(_name == null){
			throw new Exception("Application name required.");
		}
		
		if(_version == null){
			throw new Exception("Application version required.");
		}

		if(_baseVersion == null){
			throw new Exception("Application base version required.");
		}		

		DiffAAIDType aaID = objFactory.createDiffAAIDType();
		
		String name = AAID_NAME_URI + _name;
		aaID.setName(name);
		aaID.setVersion(_version);
		aaID.setBaseVersion(_baseVersion);
		return aaID;
	}

	private AuthorType createAuthor(
		String name,
		String desc,
		String country,
		String address)
		throws Exception {

		AuthorType author = objFactory.createAuthorType();

		// name
		if(name == null){
			throw new Exception("Author name required.");
		}
		author.setName(name);
		
		// description
		if(desc != null){
			DescriptionType authDesc = objFactory.createDescriptionType();
			authDesc.setValue(desc);
			author.getDescription().add(authDesc);
		}

		// set location
		if(country != null || address != null){
			
			LocationType location = objFactory.createAuthorTypeLocationType();
			
			if(country != null){
				location.setCountry(country);
			}
			
			if(address != null){
				location.setAddress(address);
			}
			
			author.setLocation(location);
		}
		
		return author;
	}


	private DescriptionsType createDescriptions(
		String _desc)
		throws JAXBException {

		DescriptionsType descs = objFactory.createDescriptionsType();
		DescriptionType desc = objFactory.createDescriptionType();
		desc.setValue(_desc);
		
		descs.getDescription().add(desc);
		return descs;
	}


	private ContentsType createContents(
		AcsFileInfo[] fileInfo)
		throws JAXBException {

		// create Contents
		ContentsType contents = objFactory.createContentsType();

		// add Content
		for (int i = 0; i < fileInfo.length; i++) {
			contents.getContent().add(
				createContent(
					fileInfo[i].getRelativeFilePath(),
					fileInfo[i].getTypeString()));
		}
		return contents;
	}

	private DiffContentsType createDiffContents(
		AcsFileInfo[] fileInfo)
		throws Exception {

		// create DiffContents
		DiffContentsType diffContents = objFactory.createDiffContentsType();

		// add Content
		for (int i = 0; i < fileInfo.length; i++) {
			diffContents.getContent().add(
				createDiffContent(
					fileInfo[i].getRelativeFilePath(),
					fileInfo[i].getTypeString(),
					fileInfo[i].getOperation()));
		}
		return diffContents;
	}

	private ContentType createContent(
		String relativePath,
		String typeString)
		throws JAXBException {

		ContentType content = objFactory.createContentType();
		if(typeString != null){
			content.setType(new QName(TYPE_NAMESPACE, typeString, TYPE_NAMESPACE_PREFIX));
		}
		content.setPathname(relativePath);

		return content;
	}

	private DiffContentType createDiffContent(
		String _relativePath,
		String _typeString,
		int _operation)
		throws Exception {

		if(AcsFileInfo.ADD != _operation && 
				AcsFileInfo.DELETE != _operation &&
				AcsFileInfo.REPLACE != _operation){
			throw new Exception("File operation (Add/Delete/Replace) required");	
		}
		
		DiffContentType content = objFactory.createDiffContentType();

		if(_typeString != null){
			content.setType(new QName(TYPE_NAMESPACE, _typeString, TYPE_NAMESPACE_PREFIX));
		}
		
		content.setPathname(_relativePath);		
		content.setOperation(AcsFileInfo.getStringFromInt(_operation));

		return content;
	}

	/**<PRE>
	 * file is read as byte[].
	 * 
	 * @param file
	 * @return
	 * @throws IOException</PRE>
	 */
	public byte[] readFileAsbyte(String file) throws IOException {

		FileInputStream in = null;
		BufferedInputStream bin = null;
		ByteArrayOutputStream out = null;
		try {

			try {
				// read file as byte
				in = new FileInputStream(file);
			} catch (FileNotFoundException e) {
				throw e;
			}

			bin = new BufferedInputStream(in);
			out = new ByteArrayOutputStream();
			byte[] buffer = new byte[1024];
			int size;

			try {
				// write to outout stream by specified buffer size
				while ((size = in.read(buffer)) != -1) {
					out.write(buffer, 0, size);

				}
				out.flush();
			} catch (IOException e) {
				throw e;
			}

			byte[] result = out.toByteArray();
			return (result);

		} finally {
			if (in != null) {
				try {
					in.close();
					out.close();
				} catch (IOException e) {
					throw e;
				}
			}
		}
	}

	/**<PRE>
	 * byte[] data is written into file.
	 * file must exist.
	 * 
	 * @param file 
	 * @param data</PRE>
	 */
	public void writeFileFrombyte(File file, byte[] data) throws IOException {

		FileOutputStream out = null;
		try {

			// file must be exist
			out = new FileOutputStream(file);

			// write byte[] data into file
			out.write(data);
			out.flush();
		} catch (FileNotFoundException e) {
			throw e;
		} catch (IOException e) {
			throw e;
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
					throw e;
				}
			}
		}
	}
}