/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2006/06/13   Fujitsu)Yositake      First Version
 */

package org.naregi.pse.acs.client;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.log4j.Logger;

import org.oasis.wsn.NotificationProducer;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.SubscribeResponse;
import org.oasis.wsn.TopicExpressionType;
import org.oasis.wsn.WSBaseNotificationServiceAddressingLocator;

import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.properties.ResourcePropertyValueChangeNotificationType;
import org.xml.sax.InputSource;

import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.NotifyCallback;

import org.globus.wsrf.WSNConstants;
import org.globus.wsrf.container.ContainerException;
import org.globus.wsrf.core.notification.ResourcePropertyValueChangeNotificationElementType;
import org.globus.wsrf.core.notification.SubscriptionManager;
import org.globus.wsrf.core.notification.service.SubscriptionManagerServiceAddressingLocator;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;

import java.io.ByteArrayInputStream;
import java.util.List;

import javax.xml.namespace.QName;

/**
 * 
 * This class will be listenning the results after the client<b> 
 * requested the acs serice.
 * 
 */
public class AcsListener implements NotifyCallback {

	private static Logger log = Logger.getLogger(AcsListener.class.getName());
	
	// topic type
	public static final int CREATED_TOPIC = 1;
	public static final int UPDATED_TOPIC = 2;
	private static final int FAULT_CREATE = 3;
	private static final int FAULT_UPDATE = 4;
	
	// 10 seconds
	private static final long WAITE_PERIOD = 10000;
	
	private EndpointReferenceType producerEPR;
	private EndpointReferenceType consumerEPR;
	
	private Looper looper;
	private NotificationConsumerManager consumerManager;	
	private SubscribeResponse normalSubscribeResponse;
	private SubscribeResponse faultSubscribeResponse;

	private boolean listening = true;
	
	public AcsListener(EndpointReferenceType epr, int opeType) throws Exception  {
		
		if((opeType != CREATED_TOPIC) && (opeType != UPDATED_TOPIC)){
		
			try {
				throw new Exception("topicType argument is neither CREATE nor UPDATE.");
			} catch (Exception e) {
				log.fatal("topicType argument is neither CREATE nor UPDATE.", e);
				throw e;
			}
		}
		this.producerEPR = epr;
		this.looper = new Looper();
		init(opeType);
	}
	
	public AcsListener(byte[] eprByte, int opeType) throws Exception  {
		
		if((opeType != CREATED_TOPIC) && (opeType != UPDATED_TOPIC)){
		
			try {
				throw new Exception("topicType argument is neither CREATE nor UPDATE.");
			} catch (Exception e) {
				log.fatal("topicType argument is neither CREATE nor UPDATE.", e);
				throw e;
			}
		}
		this.producerEPR = getEPR(eprByte);
		this.looper = new Looper();
		init(opeType);
	}

	private EndpointReferenceType getEPR(byte[] eprByte) throws DeserializationException{
		
		ByteArrayInputStream bis = new ByteArrayInputStream(eprByte);
		
		EndpointReferenceType epr =
			(EndpointReferenceType) ObjectDeserializer.deserialize(
				new InputSource(bis),
				EndpointReferenceType.class);
		
		return epr;
	}

	/**
	 * kick off listener.
	 */
	public void startListening() {
		log.info("startListening START");
		looper.start();
		log.info("startListening END");
	}

	public void stopListening() {
		this.listening = false;
	}

	/**
	 * do the initialize.
	 * @param operationType
	 * @throws Exception
	 */
	private void init(int opeType) throws Exception {
		
		try {
			
			// The NotificationConsumerManager sets up an endpoint where
			// notifications will be delivered.
			consumerManager = NotificationConsumerManager.getInstance();
			consumerManager.startListening();
			consumerEPR = consumerManager
					.createNotificationConsumer(this);

			// Create the request to the remote Subscribe() call
			Subscribe request = new Subscribe();

			// Must the notification be delivered using the Notify operation?
			request.setUseNotify(Boolean.TRUE);
			
			// Indicate what the client's EPR is
			request.setConsumerReference(consumerEPR);

			// The TopicExpression specifies what topic we want to subscribe to
			TopicExpressionType topicExpression = new TopicExpressionType();
			topicExpression.setDialect(WSNConstants.SIMPLE_TOPIC_DIALECT);
			
			// This listener subscribe for this Resource Property
			if(opeType == CREATED_TOPIC){
				topicExpression.setValue(ACSQNames.AR_RP_CREATED_MESSAGE);
			}else if(opeType == UPDATED_TOPIC){
				topicExpression.setValue(ACSQNames.AA_RP_UPDATED_MESSAGE);
			}
			
			request.setTopicExpression(topicExpression);

			// Get a reference to the NotificationProducer portType
			WSBaseNotificationServiceAddressingLocator notifLocator = 
				new WSBaseNotificationServiceAddressingLocator();
			
			// Get notification producer port type
			NotificationProducer producerPort = notifLocator
					.getNotificationProducerPort(producerEPR);

			// subscribe for normal
			normalSubscribeResponse = producerPort.subscribe(request);
			
			// subscribe for fault
			if(opeType == CREATED_TOPIC){
				topicExpression.setValue(ACSQNames.AR_RP_FAULT_MESSAGE);
			}else if(opeType == UPDATED_TOPIC){
				topicExpression.setValue(ACSQNames.AA_RP_FAULT_MESSAGE);
			}

			request.setTopicExpression(topicExpression);
			faultSubscribeResponse = producerPort.subscribe(request);

		} catch (Exception e) {
			log.fatal("Failed to subscribe.", e);
			throw e;
		}
	}


	/**
	 * This method is called when a notification is delivered
	 */	
	/* ( Javadoc)
	 * @see org.globus.wsrf.NotifyCallback#deliver(java.util.List, org.apache.axis.message.addressing.EndpointReferenceType, java.lang.Object)
	 */
	public synchronized void deliver(List topicPath, EndpointReferenceType producer,
			Object message) {
			
		int topicType = 0;

		//1. determine what notification come
		if(checkPath(ACSQNames.AR_RP_CREATED_MESSAGE, topicPath)){
			topicType = CREATED_TOPIC;
		}else if(checkPath(ACSQNames.AA_RP_UPDATED_MESSAGE, topicPath)){
			topicType = UPDATED_TOPIC;
		}else if(checkPath(ACSQNames.AR_RP_FAULT_MESSAGE, topicPath)){
			topicType = FAULT_CREATE;	
		
		// Because For this sample, AR_RP_FAULT_MESSAGE and AA_RP_FAULT_MESSAGE 
		// is same type, never come here.
		}else if(checkPath(ACSQNames.AA_RP_FAULT_MESSAGE, topicPath)){
			topicType = FAULT_UPDATE;
		}
		
		ResourcePropertyValueChangeNotificationElementType notif_elem;
		ResourcePropertyValueChangeNotificationType notif;
	
		notif_elem = (ResourcePropertyValueChangeNotificationElementType) message;
		notif = notif_elem.getResourcePropertyValueChangeNotification();
		notif.toString();
	
		if (notif != null) {
			log.info("A notification has been delivered");

			MessageElement messageElement = notif.getNewValue().get_any()[0];
			
			switch(topicType){
				
				case CREATED_TOPIC :
					log.info("Notification Type : Created Message.");

					// dump message
					log.info("Message = " + messageElement.toString());
					
					// stop Listening
					listening = false;

					break;
				
				case UPDATED_TOPIC :
					log.info("Notification Type : Updated Message.");

					// dump message
					log.info("Message = " + messageElement.toString());
										
					// stop Listening
					listening = false;

					break;
				
				case FAULT_CREATE :
					log.info("Notification Type : FaultMessage.");

					// dump message
					log.info("Message = " + messageElement.toString());

					// stop Listening
					listening = false;

					break;
				
				// never come here
				case FAULT_UPDATE :
					log.info("Notification Type : Fault from Update operation.");

					// dump message
					log.info("Message = " + messageElement.toString());

					// stop Listening
					listening = false;

					break;
			}
		}
	}

	/**
	 * check the topic path of the coming notification whether is equal<b>
	 * with the specified QName. 
	 * @param name
	 * @param topicPath
	 * @return
	 */
	private synchronized boolean checkPath(QName name, List topicPath) {
		boolean isEqual = false;
		
		Object topic = topicPath.get(0);
		if(topic instanceof QName){
			isEqual = name.equals(topic);
		}
		
		return isEqual;
	}
	
	/**
	 * 
	 * The inner class for keep the listener running.
	 */
	private class Looper extends Thread{
		
		/**
		 * execute the process.
		 */
		public void run() {
			
			//1. we will loop until one notification is delivered. 
			log.info("looper is running");
			while (listening) {
				try {
					Thread.sleep(WAITE_PERIOD);
				} catch (Exception e) {
					log.fatal("** Alert ** No Exception was returned. Interrupted while sleeping.", e);
					break;
				}
			}
				
			//2. the expected notifications have been delivered, so we stop the work.
			//   and release resources.
			try {
				
				log.info("Looping has been terminated");
				
				SubscriptionManagerServiceAddressingLocator sLocator = new SubscriptionManagerServiceAddressingLocator();
				
				// destroy for normalSubscribeResponse
				SubscriptionManager manager = sLocator.getSubscriptionManagerPort(
					normalSubscribeResponse.getSubscriptionReference());
				manager.destroy(new Destroy());
				
				// destroy for faultSubscribeResponse
				manager = sLocator.getSubscriptionManagerPort(
									faultSubscribeResponse.getSubscriptionReference());
				manager.destroy(new Destroy());
				
				log.info("Subscription has terminated.");
				
				// destroy consumer
				consumerManager.removeNotificationConsumer(consumerEPR);
				
				log.info("consumer has terminated.");
					
			}catch (Exception e) {
				log.fatal("** Alert ** No Exception was returned. Failed to destroy subscription manager resource.", e);
			}finally{
				if(consumerManager != null){
					try {
						consumerManager.stopListening();
						log.info("consumerManager has stopped listening.");
					} catch (ContainerException e) {
						log.fatal("** Alert ** No Exception was returned. Failed to stop consumer.", e);
					}
				}
				log.info("Acs Listener thread terminated.");
			}
		}				
	}
}
