/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2006/06/19   Fujitsu)Yositake      First Version 
 */

package org.naregi.pse.acs.client;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Date;

import org.apache.axis.message.MessageElement;
import org.apache.axis.types.URI;
import org.apache.log4j.Logger;

import org.naregi.pse.acs.ari.stubs.AAType;
import org.naregi.pse.acs.ari.stubs.AATypeBundle;
import org.naregi.pse.acs.ari.stubs.AATypeContent;
import org.naregi.pse.acs.ari.stubs.AATypeDescriptor;
import org.naregi.pse.acs.ari.stubs.GetArchiveResponse;
import org.naregi.pse.acs.ari.stubs.GetContentsResponse;
import org.naregi.pse.acs.ari.stubs.GetContentsResponseContent;
import org.naregi.pse.acs.client.databean.*;
import org.naregi.pse.acs.client.filemanage.FileManager;
import org.naregi.pse.acs.client.filemanage.ZipFileManager;
import org.naregi.pse.acs.client.filetransfer.FileTransfer;
import org.naregi.pse.ap.common.PSESystemInfo;
import org.w3c.dom.NodeList;

/**
 * This class provides functionalities to access ACS. 
 * 
 * @author Yositake
 */
public class ResponseFileGetter {

	private static Logger log = Logger.getLogger(ResponseFileGetter.class.getName());

	private static final String ST_TRANS_TYPE_DESCRETE =
		"http://schemas.ggf.org/acs/2006/04/ari/transport-type/discrete";

	private static final String ST_TRANS_TYPE_BUNDLED_ZIP =
		"http://schemas.ggf.org/acs/2006/04/ari/transport-type/bundled/zip";

	private static final String ST_TRANS_METHOD_EMBEDDED =
		"http://schemas.ggf.org/acs/2006/04/ari/transport-method/embedded";

	private static final String ST_TRANS_METHOD_GRIDFTP =
		"http://schemas.ggf.org/acs/2006/04/ari/transport-method/gridftp";
		
	public static final String BUNDLED_FILENAME_ZIP = "AAFile.zip";

	public static final int TRANS_TYPE_DESCRETE = 0;
	public static final int TRANS_TYPE_BUNDLED_ZIP = 1;
	public static final int TRANS_METHOD_EMBEDDED = 0;
	public static final int TRANS_METHOD_GRIDFTP = 1;
	public static final int AATYPE_FULL = 0;
	public static final int AATYPE_DIFFERENTIAL = 1;
	public static final int QUERY_DIALECT_XPATH_1_0 = 0; 

	// tmp dir
	private File tmpDir;
	
	private FileTransfer fileTransfer;
	
	private FileManager fileUtil;

	public ResponseFileGetter(){
		
		this.fileUtil = new ZipFileManager();
	}

	public void getArchiveResponse(GetArchiveResponse res) throws Throwable {
		
		AAType aa = res.getAA();
		
		// store files to tmp dir
		try{
			collectFiles(aa);
		}catch(Throwable t){
			throw t;
		}finally{
			if(this.fileTransfer != null){
			this.fileTransfer.closeConnection();
			}
		}
		
		System.out.println("File was extracted at " + this.tmpDir.getPath());
	}

	public void getContentsResponse(GetContentsResponse res) throws Throwable {

		GetContentsResponseContent[] cons = res.getContent();
		
		if(cons == null || cons.length <= 0 || cons[0] == null){
			System.out.println("No files found for the query.");
		}else{
			
			// store files to tmp dir
			try{
				collectFiles(cons);
			}catch(Throwable t){
				throw t;
			}finally{
				if(this.fileTransfer != null){
				this.fileTransfer.closeConnection();
				}
			}
			
			System.out.println("File was extracted at " + this.tmpDir.getPath());
		}
	}
	
	private void collectFiles(AAType aa) throws Exception {
		
		log.info("collectFiles START");
		
		// make tmpDir & set to property
		makeTmpDir();
		
		// get transport type
		int transType = checkTransportType(aa);
		
		// for descrete
		if(transType == ACSClient.TRANS_TYPE_DESCRETE){

			collectDescreteFiles(aa);
		
		// for bundled/zip
		}else if(transType == ACSClient.TRANS_TYPE_BUNDLED_ZIP){
			
			collectBundledZipFile(aa);
		}
		
		log.info("collectFiles END");
	}

	private void collectFiles(GetContentsResponseContent[] cons) throws Exception {
		
		log.info("collectFiles START");
		
		// make tmpDir & set to property
		makeTmpDir();
		
		collectContents(cons);
		
		log.info("collectFiles END");
	}

	private void makeTmpDir() throws Exception{
		
		log.info("makeTmpDir START");
		
		// get tmp base dir
		String strTmpDir = 
			PSESystemInfo.getInstance().getProperty(AADFileFactory.ACS_CLIENT_TEMP_BASEDIR_KEY);
		
		// check last /, \
		strTmpDir = checkDir(strTmpDir);
		
		strTmpDir = strTmpDir + "RetureFiles/" + new Date().getTime() + "/";

		File fileTmpDir = new File(strTmpDir);

		// check exists, timestamp overlap
		if(fileTmpDir.exists()){
			throw new Exception("Tmp dir already exists. " + fileTmpDir.getPath());
		}
		
		// make dir
		if(fileTmpDir.mkdirs() == false){
			throw new Exception("Faild to make dir. " + fileTmpDir.getPath());
		}
		
		// set to property
		this.tmpDir = fileTmpDir;
		
		log.info("makeTmpDir END");
	}

	/**
	 * 
	 * @param dir
	 * @return
	 */
	private String checkDir(String dir){
		
		log.info("checkDir START");
		log.info("paramDir " + dir);
		
		String retDir = dir;
		
		// attach / to last
		if(retDir.endsWith("/") == false){
			retDir = retDir + "/";
		}
		
		// swap \ to / 
		retDir = retDir.replace('\\', '/');
		
		log.info("retDir = " + retDir);
		log.info("checkDir END");
		
		return retDir ;
	}
	
	private void collectDescreteFiles(AAType aa) throws Exception{
		
		log.info("collectDescreteFiles START");
		
		// for descriptor
		AATypeDescriptor des = aa.getDescriptor();
		collectDescriptor(des);
		
		// for contents
		AATypeContent[] contents = aa.getContent();
		collectContents(contents);
		
		log.info("collectDescreteFiles END");
	}

	private void collectDescriptor(AATypeDescriptor des) throws Exception{
		
		log.info("collectDescriptor START");
		
		// get Transport Method
		int transMethod = checkTransportMethod(des.getTransportMethod());

		File descFile = null;
		String strDescFile = this.tmpDir.getPath() + "/" + AADFileFactory.AAD_FILENAME;
		descFile = new File(strDescFile);
		makeSureFileExists(descFile);
		
		// embedded
		if(transMethod == ACSClient.TRANS_METHOD_EMBEDDED){
			
			byte[] entity = des.getEmbedded();
			this.fileUtil.writeFileFrombyte(descFile, entity);
		
		// gridftp (guaranteed)
		}else{
			
			MessageElement[] any = des.get_any();
			NodeList text = any[0].getChildNodes();
			String url = text.item(0).getNodeValue();
			UrlInfo urlInfo = getUrlInfo(url);

			if(this.fileTransfer == null){
				this.fileTransfer = new FileTransfer();
			}

			fileTransfer.getFromAcsClientToAcsServer(null, urlInfo, strDescFile);
		}
		
		log.info("collectDescriptor END");
	}

	private UrlInfo getUrlInfo(String url) throws MalformedURLException{
		
		UrlInfo urlInfo = new UrlInfo();
		
		URL tmpFtpURL = null;
		if(url.startsWith("gsiftp")){
			tmpFtpURL = new URL(url.replaceFirst("gsiftp", "ftp"));
		}
		urlInfo.setAddress(tmpFtpURL.getHost());
		urlInfo.setPort(String.valueOf(tmpFtpURL.getPort()));
		urlInfo.setFile(tmpFtpURL.getFile());
		
		return urlInfo;
	}

	/**
	 * Guaranteed tmp dir is newly created for this operation.
	 * (No files exists in tmp dir.)
	 * 
	 * @param file
	 * @throws Exception
	 */
	private void makeSureFileExists(File file) throws Exception{

		// make dir inside tmp dir if needed
		File parent = file.getParentFile();
		if(parent.exists() == false){
			if(parent.mkdirs() == false){
				throw new Exception("Failed to make dir. " + parent.getPath());
			}
		}

		// create file		
		if(file.createNewFile() == false){
			throw new Exception("Failed to create file. " + file.getPath());
		}
	}
	
	private void collectContents(AATypeContent[] contents) throws Exception{
		
		log.info("collectContents START");
		
		// for each content
		for(int i = 0; i < contents.length; i++){

			// get Transport Method
			int transMethod = checkTransportMethod(contents[i].getTransportMethod());
			
			String relativePath = null;
			relativePath = checkRelativePath(contents[i].getPathname());
			String strContentFile = this.tmpDir.getPath() + "/" + relativePath;
			File contentFile = new File(strContentFile);
			makeSureFileExists(contentFile);
			
			// embedded
			if(transMethod == ACSClient.TRANS_METHOD_EMBEDDED){
				byte[] entity = contents[i].getEmbedded();
				
				// write
				this.fileUtil.writeFileFrombyte(contentFile, entity);				
		
			// gridftp (guaranteed)
			}else{
				// get
				MessageElement[] any = contents[i].get_any();
				NodeList text = any[0].getChildNodes();
				String url = text.item(0).getNodeValue();
				UrlInfo urlInfo = getUrlInfo(url);
				if(this.fileTransfer == null){
					this.fileTransfer = new FileTransfer();
				}
				fileTransfer.getFromAcsClientToAcsServer(null, urlInfo, strContentFile);
				
			}		
		}
		
		log.info("collectContents END");
	}

	private void collectContents(GetContentsResponseContent[] contents) throws Exception{
		
		log.info("collectContents START");
		
		// for each content
		for(int i = 0; i < contents.length; i++){

			// get Transport Method
			int transMethod = checkTransportMethod(contents[i].getTransportMethod());
			
			String relativePath = null;
			relativePath = checkRelativePath(contents[i].getPathname());
			String strContentFile = this.tmpDir.getPath() + "/" + relativePath;
			File contentFile = new File(strContentFile);
			makeSureFileExists(contentFile);
			
			// embedded
			if(transMethod == ACSClient.TRANS_METHOD_EMBEDDED){
				byte[] entity = contents[i].getEmbedded();
				
				// write
				this.fileUtil.writeFileFrombyte(contentFile, entity);				
		
			// gridftp (guaranteed)
			}else{
				// get
				MessageElement[] any = contents[i].get_any();
				NodeList text = any[0].getChildNodes();
				String url = text.item(0).getNodeValue();
				UrlInfo urlInfo = getUrlInfo(url);
				if(this.fileTransfer == null){
					this.fileTransfer = new FileTransfer();
				}
				fileTransfer.getFromAcsClientToAcsServer(null, urlInfo, strContentFile);
				
			}		
		}
		
		log.info("collectContents END");
	}

	private String checkRelativePath(String _relativePath){
		
		// check first /
		// check last /
		
		return _relativePath;
	}
	
	private void collectBundledZipFile(AAType aa) throws Exception{
		
		log.info("collectBundledZipFile STRAT");
		
		AATypeBundle bundle = aa.getBundle();
		
		// get Transport Method
		int transMethod = checkTransportMethod(bundle.getTransportMethod());

		String strBundledZipFile = this.tmpDir.getPath() + "/" + ACSClient.BUNDLED_FILENAME_ZIP;
		File bundledZipFile = new File(strBundledZipFile);
		makeSureFileExists(bundledZipFile);
		
		// embedded
		if(transMethod == ACSClient.TRANS_METHOD_EMBEDDED){
			
			byte[] entity = bundle.getEmbedded();
			this.fileUtil.writeFileFrombyte(bundledZipFile, entity);

		// gridftp (guaranteed)
		}else{
			// get
			MessageElement[] any = bundle.get_any();
			NodeList text = any[0].getChildNodes();
			String url = text.item(0).getNodeValue();
			UrlInfo urlInfo = getUrlInfo(url);
			if(this.fileTransfer == null){
				this.fileTransfer = new FileTransfer();
			}
			fileTransfer.getFromAcsClientToAcsServer(null, urlInfo, strBundledZipFile);
		}
				
		log.info("collectBundledZipFile END");
	}

	private int checkTransportType(AAType aa) throws Exception {
		
		URI uriTransType = aa.getTransportType();
		int intTransType = -1;
		
		// descrete
		if(uriTransType.toString().equalsIgnoreCase(ACSClient.ST_TRANS_TYPE_DESCRETE)){
			intTransType = ACSClient.TRANS_TYPE_DESCRETE;
		
		// bundled/zip
		}else if(uriTransType.toString().equalsIgnoreCase(ACSClient.ST_TRANS_TYPE_BUNDLED_ZIP)){
			intTransType = ACSClient.TRANS_TYPE_BUNDLED_ZIP;
		}
		
		return intTransType;
	}

	private int checkTransportMethod(URI uriTransMethod) throws Exception {

		int intTransMethod = -1;
		
		// embedded
		if(uriTransMethod.toString().equalsIgnoreCase(ACSClient.ST_TRANS_METHOD_EMBEDDED)){
			intTransMethod = ACSClient.TRANS_METHOD_EMBEDDED;
		
		// gridftp
		}else if(uriTransMethod.toString().equalsIgnoreCase(ACSClient.ST_TRANS_METHOD_GRIDFTP)){
			intTransMethod = ACSClient.TRANS_METHOD_GRIDFTP;
		}
		
		return intTransMethod;
	}	
}