/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2006/05/29   Fujitsu)Yositake      First Version 
 */
package org.naregi.pse.acs.service.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.apache.axis.types.URI;
import org.apache.axis.types.URI.MalformedURIException;
import org.apache.log4j.Logger;
import org.globus.wsrf.InvalidResourceKeyException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.PersistenceCallback;
import org.globus.wsrf.RemoveCallback;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceIdentifier;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.TopicListAccessor;
import org.globus.wsrf.Topic;
import org.globus.wsrf.TopicList;
import org.globus.wsrf.impl.SimpleTopicList;
import org.globus.wsrf.impl.ResourcePropertyTopic;
import org.globus.wsrf.utils.FilePersistenceHelper;
import org.globus.wsrf.utils.SubscriptionPersistenceUtils;
import org.naregi.pse.acs.aaf.stubs.AADType;
import org.naregi.pse.acs.aaf.stubs.DiffAADType;
import org.naregi.pse.acs.ari.stubs.AAInstanceStateType;
import org.naregi.pse.acs.ari.stubs.ApplicationArchiveUpdatedMessageType;
import org.naregi.pse.acs.ari.stubs.FaultMessageType;
import org.naregi.pse.acs.service.impl.constants.ACSQNames;
import org.w3c.dom.Element;

//import org.w3.www._2005._03.addressing.EndpointReferenceType;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;

/**
 * Resource class for Application Archive.
 */
public class ApplicationArchiveResource
	implements
		Resource,
		ResourceIdentifier,
		ResourceProperties,
		TopicListAccessor,
		PersistenceCallback,
		RemoveCallback {

	private static Logger log = 
				Logger.getLogger(ApplicationArchiveResource.class.getName());

	// Resource Property set
	private ResourcePropertySet propSet;

	// Resource key, uniquely identifies this resource.
	private Object key;

	// Resource properties, /ari:State
	private ResourceProperty state;

	// Resource properties, /aaf:AAD
	private ResourceProperty aad;

	// Resource properties, /aaf:DifferentialAAD
	private ResourceProperty diffAAD;

	// Resource properties, /ari:CreationDateTime
	private ResourceProperty creationDateTime;

	// Resource properties, /ari:BaseAA
	private ResourceProperty baseAA;

	// Resource properties, /ari:NewerAA
	private ResourceProperty newerAA;

	// Resource properties, /ari:Repository
	private ResourceProperty repository;

	// Resource properties, /ari:QueryExpressionDialect
	private ResourceProperty queryExpressionDialect;

	// Resource properties for notification, /ari:ApplicationArchiveUpdatedMessage
	private ResourceProperty applicationArchiveUpdated;

	// Resource properties for fault notification, /ari:FaultMessage
	private ResourceProperty faultMessage;

	// Topic list
	private TopicList topicList;

	// FilePersistenceHelper object
	private FilePersistenceHelper persistenceHelper;

	/**
	 * Initializes RPs and returns a unique identifier for this resource
	 * 
	 * @return Object(key)
	 */
	public Object initialize(
		AAInstanceStateType _state,
		AADType _aad,
		DiffAADType _diffAAD,
		String _creationDateTime,
		EndpointReferenceType _baseAA,
		EndpointReferenceType _newerAA,
		EndpointReferenceType _repository)
		throws ResourceException, MalformedURIException {

		// create key
		this.key = new String(new Date().getTime() + "" + hashCode());

		// create property set
		this.propSet =
			new SimpleResourcePropertySet(ACSQNames.AA_RESOURCE_PROPERTIES);

		// set ari:State
		this.state = new SimpleResourceProperty(ACSQNames.AA_RP_STATE);
		this.state.add(_state);

		// set aaf:AAD
		this.aad = new SimpleResourceProperty(ACSQNames.AA_RP_AAD);
		this.aad.add(_aad);

		// set aaf:DifferentialAAD
		this.diffAAD = new SimpleResourceProperty(ACSQNames.AA_RP_DIFF_AAD);
		this.diffAAD.add(_diffAAD);

		// set ari:CreationDateTime
		this.creationDateTime =
			new SimpleResourceProperty(ACSQNames.AA_RP_CREATION_DATETIME);
		this.creationDateTime.add(_creationDateTime);

		// set ari:BaseAA
		this.baseAA = new SimpleResourceProperty(ACSQNames.AA_RP_BASEAA);
		this.baseAA.add(_baseAA);

		// set ari:NewerAA
		this.newerAA = new SimpleResourceProperty(ACSQNames.AA_RP_NEWERAA);
		this.newerAA.add(_newerAA);

		// set ari:Repository
		this.repository =
			new SimpleResourceProperty(ACSQNames.AA_RP_REPOSITORY);
		repository.add(_repository);

		// set ari:QueryExpressionDialect
		this.queryExpressionDialect =
			new SimpleResourceProperty(
				ACSQNames.AA_RP_QUERY_EXPRESSION_DIALECT);
		queryExpressionDialect.add(
			new URI("http://www.w3.org/TR/1999/REC-xpath-19991116"));

		// set ari:ApplicationArchiveUpdatedMessage
		this.applicationArchiveUpdated =
			new SimpleResourceProperty(ACSQNames.AA_RP_UPDATED_MESSAGE);
		this.applicationArchiveUpdated.add(null);

		// set ari:FaultMessage
		this.faultMessage =
			new SimpleResourceProperty(ACSQNames.AA_RP_FAULT_MESSAGE);
		this.faultMessage.add(null);

		// add to topic list
		this.topicList = new SimpleTopicList(this);
		this.applicationArchiveUpdated =
			new ResourcePropertyTopic(this.applicationArchiveUpdated);
		this.faultMessage = new ResourcePropertyTopic(this.faultMessage);

		this.topicList.addTopic((Topic) this.applicationArchiveUpdated);
		this.topicList.addTopic((Topic) this.faultMessage);

		// add to property set
		this.propSet.add(this.state);
		this.propSet.add(this.aad);
		this.propSet.add(this.diffAAD);
		this.propSet.add(this.creationDateTime);
		this.propSet.add(this.baseAA);
		this.propSet.add(this.newerAA);
		this.propSet.add(this.repository);
		this.propSet.add(this.queryExpressionDialect);
		this.propSet.add(this.applicationArchiveUpdated);
		this.propSet.add(this.faultMessage);

		store();

		return this.key;
	}

	public AAInstanceStateType getState() {
		AAInstanceStateType _state = (AAInstanceStateType) this.state.get(0);
		return _state;
	}

	public AADType getAAD() {
		AADType _aad = (AADType) this.aad.get(0);
		return _aad;
	}

	public DiffAADType getDiffAAD() {
		DiffAADType _diffAAD = (DiffAADType) this.diffAAD.get(0);
		return _diffAAD;
	}

	public String getCreationDateTime() {

		String _creationDateTime = (String) this.creationDateTime.get(0);
		return _creationDateTime;
	}

	public EndpointReferenceType getBaseAA() {

		EndpointReferenceType _baseAA =
			(EndpointReferenceType) this.baseAA.get(0);
		return _baseAA;
	}

	public EndpointReferenceType[] getNewerAA() {

		EndpointReferenceType[] _newerAA =
			new EndpointReferenceType[this.newerAA.size()];
		for (int i = 0; i < _newerAA.length; i++) {
			_newerAA[i] = (EndpointReferenceType) this.newerAA.get(i);
		}
		return _newerAA;
	}

	public EndpointReferenceType getRepository(){
		
		EndpointReferenceType _repository =
			(EndpointReferenceType) this.repository.get(0);
		return _repository;
	}	

	public URI[] getQueryExpressionDialect(){
				
		URI[] _queryExpressionDialect = new URI[this.queryExpressionDialect.size()];
		for(int i = 0; i < this.queryExpressionDialect.size(); i++){
			_queryExpressionDialect[i] = (URI) this.queryExpressionDialect.get(i);
		}	

		return _queryExpressionDialect;
	}

	public ApplicationArchiveUpdatedMessageType getApplicationArchiveUpdated(){
		
		ApplicationArchiveUpdatedMessageType _updated = 
			(ApplicationArchiveUpdatedMessageType)this.applicationArchiveUpdated.get(0);
		return _updated;
		
	}

	public FaultMessageType getFaultMessage(){
		FaultMessageType _fault = (FaultMessageType)this.faultMessage.get(0);
		return _fault;
	}

	public void setState(AAInstanceStateType _state) throws ResourceException {
		this.state.set(0, _state);
		store();
	}

	public void setAAD(AADType _aad) throws ResourceException {
		this.aad.set(0, _aad);
		store();
	}

	public void setDiffAAD(DiffAADType _diffAAD) throws ResourceException {
		this.diffAAD.set(0, _diffAAD);
		store();
	}

	public void setCreationDateTime(String _creationDateTime)
		throws ResourceException {
		this.creationDateTime.set(0, _creationDateTime);
		store();
	}

	public void setBaseAA(EndpointReferenceType _baseAA)
		throws ResourceException {
		this.baseAA.set(0, _baseAA);
		store();
	}

	public void addNewerAA(EndpointReferenceType _newerAA)
		throws ResourceException {
		
		// for initialization, null is set
		if(this.newerAA.get(0) == null){
			this.newerAA.set(0, _newerAA);
		
		}else{
			this.newerAA.add(_newerAA);
		}
		store();
	}

	public void setApplicationArchiveUpdated(ApplicationArchiveUpdatedMessageType updated)
		throws ResourceException {
		this.applicationArchiveUpdated.set(0, updated);
		store();
	}

	public void setFaultMessage(Element element) throws ResourceException{
		
		FaultMessageType fault = new FaultMessageType();
		MessageElement[] me = new MessageElement[1];
		me[0] = new MessageElement(element);
		fault.set_any(me);
		this.faultMessage.set(0, fault);
		store();		
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.ResourceProperties#getResourcePropertySet()
	 * 
	 * Required by interface ResourceProperties
	 */
	public ResourcePropertySet getResourcePropertySet() {
		return this.propSet;
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.ResourceIdentifier#getID()
	 * 
	 * Required by interface ResourceIdentifier
	 */
	public Object getID() {
		return this.key;
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.TopicListAccessor#getTopicList()
	 * 
	 * Required by interface TopicListAccessor
	 */
	public TopicList getTopicList() {
		return this.topicList;
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.PersistenceCallback#store()
	 */
	public synchronized void store() throws ResourceException {
		FileOutputStream fos = null;
		File tmpFile = null;

		try {
			tmpFile =
				File.createTempFile(
					"applicationarchiveresource",
					".tmp",
					getPersistenceHelper().getStorageDirectory());
			fos = new FileOutputStream(tmpFile);
			ObjectOutputStream oos = new ObjectOutputStream(fos);

			// store State
			oos.writeObject(getState());
			
			// store AAD
			oos.writeObject(getAAD());

			// store DiffAAD
			oos.writeObject(getDiffAAD());

			// store CreationDateTime
			oos.writeObject(getCreationDateTime());

			// store BaseAA
			oos.writeObject(getBaseAA());

			// store NewerAA
			EndpointReferenceType[] _newerAA = getNewerAA();
			oos.writeInt(_newerAA.length);
			for (int i = 0; i < _newerAA.length; i++) {
				oos.writeObject(_newerAA[i]);
			}

			// store Repository
			oos.writeObject(getRepository());
			
			// store QueryExpressionDialect
			URI[] _uri = getQueryExpressionDialect();
			oos.writeInt(_uri.length);
			for(int i = 0; i < _uri.length; i++){
				oos.writeObject(_uri[i]);
			}

			// store UpdatedMessage
			oos.writeObject(getApplicationArchiveUpdated());
			
			// store FaultMessage
			oos.writeObject(getFaultMessage());

			// store SubscriptionListener
			SubscriptionPersistenceUtils.storeSubscriptionListeners(
				this.getTopicList(),
				oos);

			oos.writeObject(timestamp());

		} catch (Exception e) {
			tmpFile.delete();
			throw new ResourceException("Failed to store resource", e);
		} finally {
			if (fos != null) {
				try {
					fos.close();
				} catch (Exception e) {
					throw new ResourceException(
						"Failed to close file output stream.",
						e);
				}
			}
		}

		File file = getKeyAsFile(this.key);
		if (file.exists()) {
			file.delete();
		}
		if (!tmpFile.renameTo(file)) {
			tmpFile.delete();
			throw new ResourceException("Failed to store resource");
		}
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.PersistenceCallback#load(org.globus.wsrf.ResourceKey)
	 */
	public void load(ResourceKey key)
		throws
			ResourceException,
			NoSuchResourceException,
			InvalidResourceKeyException {

		File file = getKeyAsFile(key.getValue());

		if (!file.exists()) {
			throw new NoSuchResourceException();
		}
		initialize(key.getValue());
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(file);
			ObjectInputStream ois = new ObjectInputStream(fis);
			Object field = null;

			// load State
			field = ois.readObject();
			this.state.add(field);
			
			// load AAD
			field = ois.readObject();
			this.aad.add(field);

			// load DiffAAD
			field = ois.readObject();
			this.diffAAD.add(field);

			// load CreationDateTime
			field = ois.readObject();
			this.creationDateTime.add(field);

			// load BaseAA
			field = ois.readObject();
			this.baseAA.add(field);

			// load NewerAA
			int length = ois.readInt();
			for (int i = 0; i < length; i++) {
				field = ois.readObject();
				this.newerAA.add(field);
			}

			// load Repository
			field = ois.readObject();			
			this.repository.add(field);
			
			// load QueryExpressionDialect
			int lengthQ = ois.readInt();
			for (int i = 0; i < lengthQ; i++){
				field = ois.readObject();
				this.queryExpressionDialect.add(field);
			}

			// load UpdatedMessage
			field = ois.readObject();
			this.applicationArchiveUpdated.add(field);
			
			// load FaultMessage
			field = ois.readObject();
			this.faultMessage.add(field);
			
			// load SubscriptionListener
			SubscriptionPersistenceUtils.loadSubscriptionListeners(
				this.getTopicList(),
				ois);

		} catch (Exception e) {
			throw new ResourceException("Failed to load resource", e);
		} finally {
			if (fis != null) {
				try {
					fis.close();
				} catch (Exception e) {
					throw new ResourceException(
						"Failed to close file output stream.",
						e);
				}
			}
		}
	}

	/**
	 * Initializes only when load() is called 
	 * from FilePersistenceHelper.
	 * 
	 * @param key
	 */
	private void initialize(Object key) {

		// create key
		this.key = key;

		// create property set
		this.propSet =
			new SimpleResourcePropertySet(ACSQNames.AA_RESOURCE_PROPERTIES);

		// set ari:State
		this.state = new SimpleResourceProperty(ACSQNames.AA_RP_STATE);

		// set aaf:AAD
		this.aad = new SimpleResourceProperty(ACSQNames.AA_RP_AAD);

		// set aaf:DifferentialAAD
		this.diffAAD = new SimpleResourceProperty(ACSQNames.AA_RP_DIFF_AAD);

		// set ari:CreationDateTime
		this.creationDateTime =
			new SimpleResourceProperty(ACSQNames.AA_RP_CREATION_DATETIME);

		// set ari:BaseAA
		this.baseAA = new SimpleResourceProperty(ACSQNames.AA_RP_BASEAA);

		// set ari:NewerAA
		this.newerAA = new SimpleResourceProperty(ACSQNames.AA_RP_NEWERAA);

		// set ari:Repository
		this.repository =
			new SimpleResourceProperty(ACSQNames.AA_RP_REPOSITORY);

		// set ari:QueryExpressionDialect
		this.queryExpressionDialect =
			new SimpleResourceProperty(
				ACSQNames.AA_RP_QUERY_EXPRESSION_DIALECT);

		// set ari:ApplicationArchiveUpdatedMessage
		this.applicationArchiveUpdated =
			new SimpleResourceProperty(ACSQNames.AA_RP_UPDATED_MESSAGE);

		// set ari:FaultMessage
		this.faultMessage =
			new SimpleResourceProperty(ACSQNames.AA_RP_FAULT_MESSAGE);

		// add to topic list
		this.topicList = new SimpleTopicList(this);
		this.applicationArchiveUpdated =
			new ResourcePropertyTopic(this.applicationArchiveUpdated);
		this.faultMessage = new ResourcePropertyTopic(this.faultMessage);

		this.topicList.addTopic((Topic) this.applicationArchiveUpdated);
		this.topicList.addTopic((Topic) this.faultMessage);

		// add to property set
		this.propSet.add(this.state);
		this.propSet.add(this.aad);
		this.propSet.add(this.diffAAD);
		this.propSet.add(this.creationDateTime);
		this.propSet.add(this.baseAA);
		this.propSet.add(this.newerAA);
		this.propSet.add(this.repository);
		this.propSet.add(this.queryExpressionDialect);
		this.propSet.add(this.applicationArchiveUpdated);
		this.propSet.add(this.faultMessage);
	}

	/**
	 * get the current timestamp as a string.
	 * @return String
	 */
	private String timestamp() {

		String time =
			new SimpleDateFormat("yyyyMMddHHmmss.SSS").format(new Date());
		return "stored@" + time;
	}

	/**
	 * get a FilePersistenceHelper object.
	 * @return FilePersistenceHelper
	 */
	private synchronized FilePersistenceHelper getPersistenceHelper() {

		if (this.persistenceHelper == null) {
			try {
				this.persistenceHelper =
					new FilePersistenceHelper(getClass(), ".ser");
			} catch (Exception e) {
				throw new RuntimeException(e.getMessage());
			}
		}
		return this.persistenceHelper;
	}

	/**
	 * get key as file.
	 * @param key
	 * @return File
	 * @throws InvalidResourceKeyException
	 */
	private File getKeyAsFile(Object key) throws InvalidResourceKeyException {

		if (key instanceof String) {
			return getPersistenceHelper().getKeyAsFile(key);
		} else {
			throw new InvalidResourceKeyException();
		}
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.RemoveCallback#remove()
	 */
	public void remove() throws ResourceException {
		getPersistenceHelper().remove(this.key);

	}
}
