/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2006/05/29   Fujitsu)Yositake      First Version 
 */
package org.naregi.pse.acs.service.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.apache.axis.message.MessageElement;
import org.apache.axis.types.URI;
import org.apache.axis.types.URI.MalformedURIException;
import org.globus.wsrf.InvalidResourceKeyException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.PersistenceCallback;
import org.globus.wsrf.RemoveCallback;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceIdentifier;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.Topic;
import org.globus.wsrf.TopicList;
import org.globus.wsrf.TopicListAccessor;
import org.globus.wsrf.impl.ResourcePropertyTopic;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.impl.SimpleTopicList;
import org.globus.wsrf.utils.FilePersistenceHelper;
import org.globus.wsrf.utils.SubscriptionPersistenceUtils;
import org.naregi.pse.acs.ari.stubs.ApplicationArchiveCreatedMessageType;
import org.naregi.pse.acs.ari.stubs.FaultMessageType;
import org.naregi.pse.acs.service.impl.constants.ACSConstants;
import org.naregi.pse.acs.service.impl.constants.ACSQNames;
import org.naregi.pse.acs.service.impl.property.ACSSystemInfo;
import org.w3c.dom.Element;

/**
 * Resource class for Application Repository.
 */
public class ApplicationRepositoryResource
	implements Resource, ResourceIdentifier, ResourceProperties, TopicListAccessor, PersistenceCallback, RemoveCallback{

	// Resource Property set
	private ResourcePropertySet propSet;

	// Resource key, uniquely identifies this resource.
	private Object key;

	// Resource properties, /ari:Version
	private ResourceProperty version;

	// Resource properties, /ari:TransportType
	private ResourceProperty transportType;

	// Resource properties, /ari:TransportMethod
	private ResourceProperty transportMethod;

	// Resource properties, /ari:QueryExpressionDialect
	private ResourceProperty queryExpressionDialect;

	// Resource properties for notification, /ari:ApplicationArchiveCreatedMessage
	private ResourceProperty applicationArchiveCreated;

	// Resource properties for fault notification, /ari:FaultMessage
	private ResourceProperty faultMessage;

	// Topic list
	private TopicList topicList;

	// FilePersistenceHelper object
	private FilePersistenceHelper persistenceHelper;
	
	/**
	 * Initializes RPs and returns a unique identifier for this resource
	 * 
	 * @return Object(key)
	 * @throws MalformedURIException
	 */
	public Object initialize() throws Exception {

		// create key
		this.key = new String(ACSSystemInfo.getInstance().getProperty
								(ACSConstants.AR_RESOURCE_KEY_KEY));

		// create property set
		this.propSet =
			new SimpleResourcePropertySet(ACSQNames.AR_RESOURCE_PROPERTIES);

		// set /ari:Version
		this.version = new SimpleResourceProperty(ACSQNames.AR_RP_VERSION);
		this.version.add(new URI("http://schemas.ggf.org/acs/2006/04/ari"));

		// set /ari:TransportType
		this.transportType =
			new SimpleResourceProperty(ACSQNames.AR_RP_TRANSPORT_TYPE);
		this.transportType.add(
			new URI("http://schemas.ggf.org/acs/2006/04/ari/transport-type/discrete"));
		this.transportType.add(
			new URI("http://schemas.ggf.org/acs/2006/04/ari/transport-type/bundled/zip"));

		// set /ari:TransportMethod
		this.transportMethod =
			new SimpleResourceProperty(ACSQNames.AR_RP_TRANSPORT_METHOD);
		this.transportMethod.add(
			new URI("http://schemas.ggf.org/acs/2006/04/ari/transport-method/embedded"));
		this.transportMethod.add(
			new URI("http://schemas.ggf.org/acs/2006/04/ari/transport-method/gridftp"));

		// set /ari:QueryExpressionDialect
		this.queryExpressionDialect =
			new SimpleResourceProperty(
				ACSQNames.AR_RP_QUERY_EXPRESSION_DIALECT);
		this.queryExpressionDialect.add(
			new URI("http://www.naregi.org/namespace/acs/2006/04/lookupArchivesNotSupported"));

		// set /ari:ApplicationArchiveCreatedMessage
		this.applicationArchiveCreated =
			new SimpleResourceProperty(ACSQNames.AR_RP_CREATED_MESSAGE);
		this.applicationArchiveCreated.add(null);

		// set /ari:FaultMessage
		this.faultMessage =
			new SimpleResourceProperty(ACSQNames.AR_RP_FAULT_MESSAGE);
		this.faultMessage.add(null);

		// add to topic list
		this.topicList = new SimpleTopicList(this);
		this.applicationArchiveCreated =
			new ResourcePropertyTopic(this.applicationArchiveCreated);
		this.faultMessage = new ResourcePropertyTopic(this.faultMessage);

		this.topicList.addTopic((Topic) this.applicationArchiveCreated);
		this.topicList.addTopic((Topic) this.faultMessage);

		// add to property set
		this.propSet.add(this.version);
		this.propSet.add(this.transportType);
		this.propSet.add(this.transportMethod);
		this.propSet.add(this.queryExpressionDialect);
		this.propSet.add(this.applicationArchiveCreated);
		this.propSet.add(this.faultMessage);

		store();
		
		return key;
	}

	public void setApplicationArchiveCreated(ApplicationArchiveCreatedMessageType created) throws ResourceException {
		this.applicationArchiveCreated.set(0, created);
		store();
	}

	public void setFaultMessage(Element element) throws ResourceException {
		
		FaultMessageType fault = new FaultMessageType();
		MessageElement[] me = new MessageElement[1];
		me[0] = new MessageElement(element);
		fault.set_any(me);
		this.faultMessage.set(0, fault);
		store();
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.ResourceProperties#getResourcePropertySet()
	 * 
	 * Required by interface ResourceProperties
	 */
	public ResourcePropertySet getResourcePropertySet() {
		return this.propSet;
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.ResourceIdentifier#getID()
	 * 
	 * Required by interface ResourceIdentifier
	 */
	public Object getID() {
		return this.key;
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.TopicListAccessor#getTopicList()
	 */
	public TopicList getTopicList() {
		return this.topicList;
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.PersistenceCallback#load(org.globus.wsrf.ResourceKey)
	 */
	public void load(ResourceKey key) throws ResourceException, NoSuchResourceException, InvalidResourceKeyException {

		File file = getKeyAsFile(key.getValue());

		if (!file.exists()) {
			throw new NoSuchResourceException();
		}
		try {
			initialize(key.getValue());
		} catch (Exception e) {
			if(e.getMessage() != null){
				throw new ResourceException(e.getMessage());
			}else{
				throw new ResourceException();
			}
		}
		
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(file);
			ObjectInputStream ois = new ObjectInputStream(fis);
			Object field = null;

			// load /ari:Version
			field = ois.readObject();
			this.version.add(field);
			
			// load /ari:TransportType
			int lengthTT = ois.readInt();
			for (int i = 0; i < lengthTT; i++){
				field = ois.readObject();
				this.transportType.add(field);
			}

			// load /ari:TransportMethod
			int lengthTM = ois.readInt();
			for (int i = 0; i < lengthTM; i++){
				field = ois.readObject();
				this.transportMethod.add(field);
			}
			
			// load /ari:QueryExpressionDialect
			int lengthQ = ois.readInt();
			for (int i = 0; i < lengthQ; i++){
				field = ois.readObject();
				this.queryExpressionDialect.add(field);
			}

			// load /ari:ApplicationArchiveCreatedMessage
			field = ois.readObject();
			this.applicationArchiveCreated.add(field);
			
			// load /ari:FaultMessage
			field = ois.readObject();
			this.faultMessage.add(field);
			
			// load SubscriptionListener
			SubscriptionPersistenceUtils.loadSubscriptionListeners(
				this.getTopicList(),
				ois);

		} catch (Exception e) {
			throw new ResourceException("Failed to load resource", e);
		} finally {
			if (fis != null) {
				try {
					fis.close();
				} catch (Exception e) {
					throw new ResourceException(
						"Failed to close file output stream.",
						e);
				}
			}
		}		
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.PersistenceCallback#store()
	 */
	public void store() throws ResourceException {

		FileOutputStream fos = null;
		File tmpFile = null;

		try {
			tmpFile =
				File.createTempFile(
					"applicationrepositoryresource",
					".tmp",
					getPersistenceHelper().getStorageDirectory());
			fos = new FileOutputStream(tmpFile);
			ObjectOutputStream oos = new ObjectOutputStream(fos);

			// store /ari:Version
			oos.writeObject(getVersion());
			
			// store /ari:TransportType
			URI[] _uriTraType = getTransportType();
			oos.writeInt(_uriTraType.length);
			for(int i = 0; i < _uriTraType.length; i++){
				oos.writeObject(_uriTraType[i]);
			}			

			// store /ari:TransportMethod
			URI[] _uriTraMethod = getTransportMethod();
			oos.writeInt(_uriTraMethod.length);
			for(int i = 0; i < _uriTraMethod.length; i++){
				oos.writeObject(_uriTraMethod[i]);
			}
			
			// store QueryExpressionDialect
			URI[] _uriQuery = getQueryExpressionDialect();
			oos.writeInt(_uriQuery.length);
			for(int i = 0; i < _uriQuery.length; i++){
				oos.writeObject(_uriQuery[i]);
			}

			// store /ari:ApplicationArchiveCreatedMessage
			oos.writeObject(getApplicationArchiveCreated());
			
			// store /ari:FaultMessage
			oos.writeObject(getFaultMessage());

			// store SubscriptionListener
			SubscriptionPersistenceUtils.storeSubscriptionListeners(
				this.getTopicList(),
				oos);

			oos.writeObject(timestamp());

		} catch (Exception e) {
			tmpFile.delete();
			throw new ResourceException("Failed to store resource", e);
		} finally {
			if (fos != null) {
				try {
					fos.close();
				} catch (Exception e) {
					throw new ResourceException(
						"Failed to close file output stream.",
						e);
				}
			}
		}

		File file = getKeyAsFile(this.key);
		if (file.exists()) {
			file.delete();
		}
		if (!tmpFile.renameTo(file)) {
			tmpFile.delete();
			throw new ResourceException("Failed to store resource");
		}		
	}

	/* ( Javadoc)
	 * @see org.globus.wsrf.RemoveCallback#remove()
	 */
	public void remove() throws ResourceException {

		getPersistenceHelper().remove(this.key);
		
	}

	/**
	 * get the current timestamp as a string.
	 * @return String
	 */
	private String timestamp() {

		String time =
			new SimpleDateFormat("yyyyMMddHHmmss.SSS").format(new Date());
		return "stored@" + time;
	}

	/**
	 * get a FilePersistenceHelper object.
	 * @return FilePersistenceHelper
	 */
	private synchronized FilePersistenceHelper getPersistenceHelper() {

		if (this.persistenceHelper == null) {
			try {
				this.persistenceHelper =
					new FilePersistenceHelper(getClass(), ".ser");
			} catch (Exception e) {
				throw new RuntimeException(e.getMessage());
			}
		}
		return this.persistenceHelper;
	}

	/**
	 * get key as file.
	 * @param key
	 * @return File
	 * @throws InvalidResourceKeyException
	 */
	private File getKeyAsFile(Object key) throws InvalidResourceKeyException {

		if (key instanceof String) {
			return getPersistenceHelper().getKeyAsFile(key);
		} else {
			throw new InvalidResourceKeyException();
		}
	}
	/**
	 * @return
	 */
	public ApplicationArchiveCreatedMessageType getApplicationArchiveCreated() {
		ApplicationArchiveCreatedMessageType _created = 
			(ApplicationArchiveCreatedMessageType)this.applicationArchiveCreated.get(0);
		return _created;
	}

	/**
	 * @return
	 */
	public FaultMessageType getFaultMessage() {
		FaultMessageType _fault = (FaultMessageType)this.faultMessage.get(0);
		return _fault;
	}

	/**
	 * @return
	 */
	public URI[] getQueryExpressionDialect() {
		
		URI[] _queryExpressionDialect = new URI[this.queryExpressionDialect.size()];
		
		for(int i = 0; i < this.queryExpressionDialect.size(); i++){
			_queryExpressionDialect[i] = (URI) this.queryExpressionDialect.get(i);
		}	

		return _queryExpressionDialect;
	}

	/**
	 * @return
	 */
	public URI[] getTransportMethod() {

		URI[] _transportMethod = new URI[this.transportMethod.size()];
		
		for(int i = 0; i < this.transportMethod.size(); i++){
			_transportMethod[i] = (URI) this.transportMethod.get(i);
		}	

		return _transportMethod;
	}

	/**
	 * @return
	 */
	public URI[] getTransportType() {
		
		URI[] _transportType = new URI[this.transportType.size()];
		
		for(int i = 0; i < this.transportType.size(); i++){
			_transportType[i] = (URI) this.transportType.get(i);
		}	

		return _transportType;
	}

	/**
	 * @return
	 */
	public URI getVersion() {
		return (URI)this.version.get(0);
	}

	public void initialize(Object key) throws MalformedURIException, ResourceException {

		// create key
		this.key = key;

		// create property set
		this.propSet =
			new SimpleResourcePropertySet(ACSQNames.AR_RESOURCE_PROPERTIES);

		// set /ari:Version
		this.version = new SimpleResourceProperty(ACSQNames.AR_RP_VERSION);

		// set /ari:TransportType
		this.transportType =
			new SimpleResourceProperty(ACSQNames.AR_RP_TRANSPORT_TYPE);

		// set /ari:TransportMethod
		this.transportMethod =
			new SimpleResourceProperty(ACSQNames.AR_RP_TRANSPORT_METHOD);

		// set /ari:QueryExpressionDialect
		this.queryExpressionDialect =
			new SimpleResourceProperty(
				ACSQNames.AR_RP_QUERY_EXPRESSION_DIALECT);

		// set /ari:ApplicationArchiveCreatedMessage
		this.applicationArchiveCreated =
			new SimpleResourceProperty(ACSQNames.AR_RP_CREATED_MESSAGE);

		// set /ari:FaultMessage
		this.faultMessage =
			new SimpleResourceProperty(ACSQNames.AR_RP_FAULT_MESSAGE);

		// add to topic list
		this.topicList = new SimpleTopicList(this);
		this.applicationArchiveCreated =
			new ResourcePropertyTopic(this.applicationArchiveCreated);
		this.faultMessage = new ResourcePropertyTopic(this.faultMessage);

		this.topicList.addTopic((Topic) this.applicationArchiveCreated);
		this.topicList.addTopic((Topic) this.faultMessage);

		// add to property set
		this.propSet.add(this.version);
		this.propSet.add(this.transportType);
		this.propSet.add(this.transportMethod);
		this.propSet.add(this.queryExpressionDialect);
		this.propSet.add(this.applicationArchiveCreated);
		this.propSet.add(this.faultMessage);
	}
}
