/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2006/06/12   Fujitsu)Yositake      First Version 
 */
 
package org.naregi.pse.acs.service.impl;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerException;

import org.apache.axis.message.MessageElement;
import org.apache.axis.types.URI;
import org.apache.log4j.Logger;
import org.apache.xpath.XPathAPI;
import org.naregi.pse.acs.aaf.stubs.AADType;
import org.naregi.pse.acs.ari.stubs.GetContents;
import org.naregi.pse.acs.ari.stubs.GetContentsResponse;
import org.naregi.pse.acs.ari.stubs.GetContentsResponseContent;
import org.naregi.pse.acs.ari.stubs.QueryExpressionType;
import org.naregi.pse.acs.service.impl.constants.ACSConstants;
import org.naregi.pse.acs.service.impl.constants.ACSQNames;
import org.naregi.pse.acs.service.impl.exception.InvalidQueryExpressionException;
import org.naregi.pse.acs.service.impl.exception.UnknownQueryDialectException;
import org.naregi.pse.acs.service.impl.functionbean.AADCheck;
import org.naregi.pse.acs.service.impl.functionbean.AADTransformerForJaxb;
import org.naregi.pse.acs.service.impl.functionbean.AcsFileManager;
import org.naregi.pse.acs.service.impl.functionbean.AnyHelper;
import org.naregi.pse.acs.service.impl.property.ACSSystemInfo;
import org.naregi.pse.acs.service.impl.functionbean.ZipFileManager;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * 
 * To implement the GetContents operation.
 * 
 */
public class GetContentsOperation extends CommonOperation {

	private static Logger log = Logger.getLogger(GetContentsOperation.class.getName());
	
	private ApplicationArchiveResource aaResource;
	
	private GetContents request;
	
	private String gridFtpAddress;
	
	private String gridFtpPort;
	
	/**
	 * Constructor
	 * 
	 * @param _aaResource
	 * @param _request
	 * @throws Exception
	 */
	public GetContentsOperation(ApplicationArchiveResource _aaResource, GetContents _request) throws Exception{
		
		this.aaResource = _aaResource;
		this.request = _request;
		
		this.fileUtil = new AcsFileManager();
		
		this.gridFtpAddress = ACSSystemInfo.getInstance().getProperty(ACSConstants.ACS_SERVICE_GRIDFTP_ADDRESS_KEY);
		
		this.gridFtpPort = ACSSystemInfo.getInstance().getProperty(ACSConstants.ACS_SERVICE_GRIDFTP_PORT_KEY);
	}	

	/**
	 * GetContents
	 * @return
	 * @throws Exception
	 */
	public GetContentsResponse getContents() throws Exception{

		log.info("getContentsOperation START");

		// get XPath and set relativeFileList
		NodeList list = evaluateXPath();
		
		// set relativeFileList
		setRelativePath(list);
		
		// create tmp dir
		makeTmpDir();

		// prepare files
		prepareFiles();
		
		// create response
		GetContentsResponse res = createResponse();		

		log.info("getContentsOperation END");
		return res;
	}

	private NodeList evaluateXPath() throws Exception{
		
		log.info("evaluateXPath START");
		
		// get QueryExpressionType
		QueryExpressionType queryExpression = this.request.getQueryExpression();
		
		// get dialog
		String dialect = queryExpression.getDialect().toString();
		
		// check dialect supported
		checkSupported(dialect);
		
		// get query
		MessageElement[] query = queryExpression.get_any();
		String[] queryArray = AnyHelper.toString(query);
		
		// get AAD
		AADType aad = this.aaResource.getAAD();
		
		// to String
		AADTransformerForJaxb trans = new AADTransformerForJaxb();
		trans.gt4AAD2JaxbAAD(aad);
		String strAAD = trans.getJaxbAADAsString();
		
		// do query. Currently, query restricted to one
		NodeList list = doQuery(strAAD, queryArray[0]);
		
		log.info("evaluateXPath END");
		
		return list;
	}

	private void checkSupported(String dialect) throws UnknownQueryDialectException{
		
		boolean found = false;
		URI[] supported = this.aaResource.getQueryExpressionDialect();
		
		for(int i = 0; i < supported.length; i++){
			if(supported[i].toString().equalsIgnoreCase(dialect)){
				found = true;
			}
		}
		
		if(found == false){
			throw new UnknownQueryDialectException("dialect = " + dialect + " not supported." );
		}
	}

	private NodeList doQuery(String strAAD, String strQuery) throws Exception{

		log.info("doQuery START");
		log.info("AAD = " + strAAD);
		log.info("query = " + strQuery);
		
		// make document
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		
		// config namespace
		factory.setNamespaceAware(true);
		DocumentBuilder builder = factory.newDocumentBuilder();
		ByteArrayInputStream bis = new ByteArrayInputStream(strAAD.getBytes());
		Document doc = builder.parse(new InputSource(bis));
				
        // do query
		NodeList content = null;
		try {
			content = XPathAPI.selectNodeList(doc, strQuery);
		} catch (TransformerException e) {
			log.fatal("local error ", e);
			if(e.getMessage() == null){
				throw new InvalidQueryExpressionException("Failed to query = " + strQuery + 
					" Check expression again.");
			}else{
				throw new InvalidQueryExpressionException("Failed to query = " + strQuery + 
					" Check expression again. " + e.getMessage());
			}
		}

		log.info("doQuery END");
		return content;
	}

	private void setRelativePath(NodeList list){
		log.info("setRelativePath START");
		
		this.relativeFileList = new ArrayList();
		
		NodeList path = null;
		if(list != null){
			
			log.info("NodeList length = " + list.getLength());
			for(int i = 0; i < list.getLength(); i++){
				
				if(list.item(i) != null){
				
					log.info("Local name of result = " + list.item(i).getLocalName());
					path = ((Element)list.item(i)).getElementsByTagNameNS(ACSQNames.AAF, "Pathname");
			
					if(path != null && path.item(0) != null){
				
						// Pathname is no more than 1 entry
						this.relativeFileList.add(path.item(0).getFirstChild().getNodeValue());
						log.info("relativeFilePath = " + path.item(0).getFirstChild().getNodeValue());
					}else{
						log.info("/AAD/Contents/Content/Pathname not found.");
					}
				}
			}	
		}else{
			log.info("Query output NodeList is null.");
		}
		log.info("setRelativePath END");
	}

	private GetContentsResponse createResponse() throws Exception{

		// create response message
		GetContentsResponse res = new GetContentsResponse();
		
		// get transport method
		int transMethod = checkTransportMethod(this.request.getTransportMethod());

		// set
		res.setContent(getContent(transMethod));
		
		return res;
	}

	private GetContentsResponseContent[] getContent(
		int transMethod)
		throws Exception {

		// set content
		GetContentsResponseContent[] contents = 
				new GetContentsResponseContent[this.relativeFileList.size()];
		
		for(int i = 0; i < contents.length; i++){
			GetContentsResponseContent content = new GetContentsResponseContent();
			
			// for Embedded
			if (transMethod == ACSConstants.TRANS_METHOD_EMBEDDED) {
				content.setTransportMethod(new URI(ACSConstants.ST_TRANS_METHOD_EMBEDDED));
				content.setEmbedded(this.fileUtil.readFileAsbyte(this.tmpDir.getPath() + "/" + (String)this.relativeFileList.get(i)));
				content.setPathname((String)this.relativeFileList.get(i));
			
			// for Gridftp
			}else if (transMethod == ACSConstants.TRANS_METHOD_GRIDFTP) {
				content.setTransportMethod(new URI(ACSConstants.ST_TRANS_METHOD_GRIDFTP));
				MessageElement[] any = new MessageElement[1];
				any[0] = new MessageElement(ACSQNames.URL);
				any[0].addTextNode("gsiftp://" + this.gridFtpAddress + ":" + this.gridFtpPort + this.tmpDir.getPath() + "/" + (String)this.relativeFileList.get(i));
				content.set_any(any);
				content.setPathname((String)this.relativeFileList.get(i));
			}
			
			// set content 
			contents[i] = content;
		
		}// end of for loop
		
		return contents;
	}

	/**
	 * candidate of Common
	 * @throws Exception
	 */
	private void prepareFiles() throws Exception{
		
		// get AA path in repository
		File aaStoredDir = ((AcsFileManager)this.fileUtil).getAAStoredDirFromAAID
				(new AADCheck().getRepoBaseDir(), this.aaResource.getAAD().getAAID());
		
		// copy the AA to tmp
		
		// get AA file path
		File aaFile = new File(aaStoredDir.getPath() + "/" + ACSConstants.BUNDLED_FILENAME_ZIP);
	
		// copy AA file to tmp dir
		((AcsFileManager)this.fileUtil).copyBaseFileToTmpDir(aaFile, this.tmpDir);
		
		// unzip AA file in tmp dir
		((ZipFileManager)this.fileUtil).unZipWithDir
					(this.tmpDir, 
					new File(this.tmpDir.getPath() + "/" + ACSConstants.BUNDLED_FILENAME_ZIP));		
	}
}
