/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2006/06/06   Fujitsu)Yositake      First Version 
 */

package org.naregi.pse.acs.service.impl.functionbean;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import org.apache.log4j.Logger;

import org.naregi.pse.acs.service.impl.constants.ACSConstants;
import org.naregi.pse.acs.service.impl.exception.IllegalDescriptorException;
import org.naregi.pse.acs.service.impl.jaxb.AAD;
import org.naregi.pse.acs.service.impl.jaxb.ContentType;
import org.naregi.pse.acs.service.impl.jaxb.ContentsType;
import org.naregi.pse.acs.service.impl.jaxb.DiffContentType;
import org.naregi.pse.acs.service.impl.jaxb.DiffContentsType;
import org.naregi.pse.acs.service.impl.jaxb.DifferentialAAD;
import org.naregi.pse.acs.service.impl.property.ACSSystemInfo;

public class AADCheck {

	private static Logger log = Logger.getLogger(AADCheck.class.getName());
	
	private AAD jaxbAAD;
	
	private DifferentialAAD jaxbDiffAAD;
	
	private File repoBaseDir; 

	public AADCheck() throws Exception{
		
		// get repository base dir from property file
		String strRepoBaseDir = 
			ACSSystemInfo.getInstance().getProperty(ACSConstants.ACS_REPOSITORY_BASEDIR_KEY);
		
		this.repoBaseDir = new File(strRepoBaseDir);
	}
	

	public void checkAAD(String desc, ArrayList relativeFileList) throws Exception{
		
		// check if conform to AAF
		checkGrammerAAD(desc);

		// check file structure
		checkFileStructureAAD(relativeFileList);
		
		// check required
		checkRequiredAAD();
		
	}
	
	public void checkDiffAAD(String desc, ArrayList relativeFileList) throws Exception{
		
		// check if conform to AAF
		checkGrammerDiffAAD(desc);

		// check file structure
		checkFileStructureDiffAAD(relativeFileList);
		
		// check required
		checkRequiredDiffAAD();
		
	}

	private void checkRequiredAAD() throws Exception{
		
		// check name
		if(this.jaxbAAD.getAAID().getName() == null || this.jaxbAAD.getAAID().getName().equals("")){
			throw new Exception("/AAD/AAID/Name element can't be empty for this implementation.");
		}
		
		// check version
		if(this.jaxbAAD.getAAID().getVersion() == null || this.jaxbAAD.getAAID().getVersion().equals("")){
			throw new Exception("/AAD/AAID/Version element can't be empty for this implementation.");
		}
		
		// check author
		if(this.jaxbAAD.getAuthor().getName() == null || this.jaxbAAD.getAuthor().getName().equals("")){
			throw new Exception("/AAD/Author/Name element can't be empty for this implementation.");
		}
	}

	private void checkRequiredDiffAAD() throws Exception{
		
		// check name
		if(this.jaxbDiffAAD.getAAID().getName() == null || this.jaxbDiffAAD.getAAID().getName().equals("")){
			throw new Exception("/DifferentialAAD/AAID/Name element can't be empty for this implementation.");
		}
		
		// check new version
		if(this.jaxbDiffAAD.getAAID().getVersion() == null || this.jaxbDiffAAD.getAAID().getVersion().equals("")){
			throw new Exception("/DifferentialAAD/AAID/Version element can't be empty for this implementation.");
		}		
		
		// check base version
		if(this.jaxbDiffAAD.getAAID().getBaseVersion() == null || this.jaxbDiffAAD.getAAID().getBaseVersion().equals("")){
			throw new Exception("/DifferentialAAD/AAID/BaseVersion element can't be empty for this implementation.");
		}
	}

	private void checkGrammerAAD(String desc) throws Exception{

		log.info("checkGrammer START");

		// get context
		JAXBContext jaxbContext = JAXBContext
					.newInstance("org.naregi.pse.acs.service.impl.jaxb");

		// create Unmarshaller
		Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();

		// set for validation enabled unmarshall
		unmarshaller.setValidating(true);

		// unmarshal
		try {
			this.jaxbAAD = (AAD) unmarshaller
						.unmarshal(new FileInputStream(desc));
		} catch (ClassCastException e){
			log.fatal("local error ", e);
			throw new IllegalDescriptorException
					("Not DifferentialAAD But AAD must be supplied for Create operation.");
			
		} catch (JAXBException e) {
			if(e.getMessage() == null){
				throw new IllegalDescriptorException("AAD might not conform to AAF specification.");
			}else{
				throw new IllegalDescriptorException
				("AAD might not conform to AAF specification. " + e.getMessage());
			}
		}

		log.info("checkGrammer END");
	}

	private void checkGrammerDiffAAD(String desc) throws Exception{

		log.info("checkGrammerDiffAAD START");

		// get context
		JAXBContext jaxbContext = JAXBContext
					.newInstance("org.naregi.pse.acs.service.impl.jaxb");

		// create Unmarshaller
		Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();

		// set for validation enabled unmarshall
		unmarshaller.setValidating(true);

		// unmarshal
		try{
			this.jaxbDiffAAD = (DifferentialAAD) unmarshaller
					.unmarshal(new FileInputStream(desc));
	
		} catch (ClassCastException e){
			log.fatal("local error ", e);
			throw new IllegalDescriptorException
					("Not AAD But DifferentialAAD must be supplied for Update operation.");
		
		} catch (JAXBException e) {
			log.fatal("local error ", e);
			if(e.getMessage() == null){
				throw new IllegalDescriptorException("Differential AAD might not conform to AAF specification.");
			}else{
				throw new IllegalDescriptorException
				("Differential AAD might not conform to AAF specification. " + e.getMessage());
			}
		}

		log.info("checkGrammerDiffAAD END");
	}

	/**
	 * It checks if files specified in AAD are in phisical files.
	 * It does not check if phisical files are in AAD.
	 * 
	 * @param relativeFileList - physical file list
	 */
	private void checkFileStructureAAD(ArrayList relativeFileList) throws IllegalDescriptorException{

		log.info("checkFileStructureAAD START");
		
		ContentsType contents = this.jaxbAAD.getContents();
		List list = contents.getContent();
		ContentType content = null;
		
		// loop for content.
		for(int i = 0; i < list.size(); i++){
			
			// get ContentType
			content = (ContentType)list.get(i);
			log.info("contentPathname = " + content.getPathname());
			
			// match physical file list
			boolean found = false;
			for(int j = 0; j < relativeFileList.size(); j++){
				
				if(content.getPathname().equalsIgnoreCase((String)(relativeFileList.get(j)))){
					found = true;
					break;
				}
			}
			
			// if not found
			if(found == false){
				throw new IllegalDescriptorException
				(content.getPathname() + " in AAD not found as physical file.");
			}
		}

		log.info("checkFileStructureAAD END");
	}

	/**
	 * It checks if files specified in AAD are in phisical files.
	 * It does not check if phisical files are in AAD.
	 * 
	 * @param relativeFileList - physical file list
	 */
	private void checkFileStructureDiffAAD(ArrayList relativeFileList) throws IllegalDescriptorException{

		log.info("checkFileStructureDiffAAD START");

		DiffContentsType contents = this.jaxbDiffAAD.getContents();
		List list = contents.getContent();
		DiffContentType content = null;
		
		// loop for content.
		for(int i = 0; i < list.size(); i++){
			
			// get ContentType
			content = (DiffContentType)list.get(i);
			log.info("contentPathname = " + content.getPathname());
			log.info("contentOperation = " + content.getOperation());
			
			// if operation is delete, physical file not needed
			if(content.getOperation().equals("delete")){
				continue;
			}
			
			// match physical file list
			boolean found = false;
			for(int j = 0; j < relativeFileList.size(); j++){
				
				if(content.getPathname().equalsIgnoreCase((String)(relativeFileList.get(j)))){
					found = true;
					break;
				}
			}
			
			// if not found
			if(found == false){
				throw new IllegalDescriptorException
				(content.getPathname() + " in DiffAAD not found as physical file. " +
					"\n operation = " + content.getOperation());
			}
		}		

		log.info("checkFileStructureDiffAAD START");
	}


	/**
	 * @return
	 */
	public AAD getJaxbAAD() {
		return jaxbAAD;
	}


	/**
	 * @return
	 */
	public File getRepoBaseDir() {
		return repoBaseDir;
	}

	/**
	 * @return
	 */
	public DifferentialAAD getJaxbDiffAAD() {
		return jaxbDiffAAD;
	}
}
