/*
 * Copyright 2004-2006 National Institute of Informatics in Japan, 
 * All rights reserved.
 *
 * This file or a portion of this file is licensed under the terms of
 * the NAREGI Public License, found at http://www.naregi.org/download/index.html.
 * If you redistribute this file, with or without modifications, you must 
 * include this notice in the file.
 *
 * [History]
 * Date         Name                  Contents
 * ------------------------------------------------------------------------
 * 2005/09/13   Fujitsu)Koga          First Version 
 */
package org.naregi.pse.ap.batch.acscleaner;

import java.net.BindException;
import java.net.ServerSocket;
import java.net.Socket;

import org.apache.log4j.Logger;
import org.naregi.pse.acs.service.impl.property.ACSSystemInfo;

/**
 * @author Koga
 *
 * 
 * 
 */
public class ACSStartFileManagerBatch {

	// MonitoringTime (Default)
	private long monitoringTime = 10000;

	private static Logger log = Logger.getLogger(ACSStartFileManagerBatch.class.getName());
	
	/**
	 * Constructor
	 *
	 */
	public ACSStartFileManagerBatch() throws Exception {
		// Set monitoring time.
		String monitoringTimeStr = ACSSystemInfo.getInstance().getProperty("MONITORING_TIME");
		this.monitoringTime = new Long(monitoringTimeStr).longValue();
	}

	private void startBatchProgram(StopSignalReceiver rcv, ACSTempFileDeleteManager manager) {
		// Operating processing is repeated until stop directions come out.
		while(rcv.nonstop){
			Thread work = new Batch(manager);
			work.start();
			try{
				work.join();
				// Waiting ...
				Thread.sleep(this.monitoringTime);
			} catch (InterruptedException e){
				log.fatal("Interruption occurred at the time of bussiness processing execution.",e);
				log.fatal("Processing is interrupted.");
				break;
			}
		}
		log.info("Processing end.");
	}
		
	/**
	 * Main method
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		try{
			
			ACSStartFileManagerBatch batch = new ACSStartFileManagerBatch();
			//get the PSEFileManager
			ACSTempFileDeleteManager manager = ACSTempFileDeleteManager.getInstance();
			
			// The program for stop signal reception
			StopSignalReceiver rcv = StopSignalReceiver.getInstance();
			// A port number is set when a port number is passed by the argument.
			if(args.length==1){
				rcv.setPortNumber( Integer.parseInt(args[0]));
			}
			// Stop signal surveillance start
			rcv.start();
			
			// Start Bussiness Processing (This processing repeats)
			batch.startBatchProgram(rcv,manager);
			
		} catch(Exception e){
			log.fatal("Exception", e);
		}

	}
	
}

/**
 * This is Thread Class.
 * (Singlton Class)
 * 
 * @author Fujitsu)koga
 */
class StopSignalReceiver extends Thread {

	private static Logger log = Logger.getLogger(StopSignalReceiver.class.getName());

	// Port Number for Socket (Default)
	public static final int PORT_NUMBER = 9000;
	// Non Stop Flag (If the flag is "true", this thread will not stop.)
	public boolean nonstop = true;
	// Port Number
	private int portNumber;
	// 
	private static StopSignalReceiver myself = null;

	/**
	 * Constructor
	 */
	private StopSignalReceiver() {
		// Set default port number
		portNumber = PORT_NUMBER;
	}

	/*
	 * This is method to get Instance.
	 * 
	 * @return StopSignalReceiver
	 */
	public static StopSignalReceiver getInstance() {
		if (myself == null) {
			myself = new StopSignalReceiver();
		}
		return myself;
	}

	/**
	 * Set the port number.
	 * 
	 * @param inPortNumber
	 */
	public void setPortNumber(int inPortNumber) {
		portNumber = inPortNumber;
	}

	/**
	 * This is method to run the thread.
	 */
	public void run() {
		try {
			// Create Server in the defined port.
			ServerSocket waitingSocket = new ServerSocket(portNumber);
			try {
				// It stands by until there is access from a stop program.
				Socket workSocket = waitingSocket.accept();
				// Change flag to "false". (Stop Signal)
				nonstop = false;

			} catch (Exception e) {
				throw e;
			} finally {
				waitingSocket.close();
			}

		} catch (BindException e) {
			log.fatal(
				"Or more 2 surveillance system cannot be started simultaneously.",e);
			// System forced termination
			System.exit(0);
		} catch (Exception e) {
			log.fatal(
				"Since the exception occurred in the waiting for a stop signal, a surveillance system is forced to terminate.",e);
			// System forced termination
			System.exit(0);
		}
	}
}

/**
 * This is Thread Class for Bussiness processing.
 * 
 * @author Fujitsu)koga
 *
 */
class Batch extends Thread {
	
	private static Logger log = Logger.getLogger(Batch.class.getName());
	
	private ACSTempFileDeleteManager fileManager = null;
	
	/**
	 * Constructor
	 * @param manager
	 */
	public Batch(ACSTempFileDeleteManager manager){
		this.fileManager = manager;
	}
	
	/**
	 * Run method
	 */
	public void run() {
		log.info("ACSTempFileDeleteBatch#run START");
		
		try {
			// Execute the Bussiness processing
			exeProc();
		} catch (Exception e) {
			log.fatal("Exception", e);
		}
		
		log.info("ACSTempFileDeleteBatch#run END");
	}

	/**
	 * Execute the synchronized method.
	 *
	 */
	private synchronized void exeProc(){
		
		fileManager.deleteTmpDir();
		
	}
}
